from heapqueue import BaseHeap


class BinaryHeap(BaseHeap):
    def __init__(self, **kwargs):
        super(BinaryHeap, self).__init__()
        self.comp = kwargs['comp'] if 'comp' in kwargs else lambda a, b: a - b
        self.size = kwargs['size'] if 'size' in kwargs else -1
        self.__heap = [None]
        self.__n = 0

    def _swap(self, i, j):
        self.__heap[i], self.__heap[j] = self.__heap[j], self.__heap[i]

    def _up(self, k):
        while k > 1 and self.comp(self.__heap[k], self.__heap[k // 2]) < 0:
            self._swap(k // 2, k)
            k //= 2

    def _down(self, k):
        while k << 1 <= self.__n:
            j = k << 1
            if j < self.__n and self.comp(self.__heap[j + 1],
                                          self.__heap[j]) < 0:
                j += 1
            if not self.comp(self.__heap[j], self.__heap[k]) < 0:
                break
            self._swap(k, j)
            k = j

    def push(self, val):
        self.__heap.append(val)
        self.__n += 1
        self._up(self.__n)

    def peek(self):
        return None if self.empty else self.__heap[1]

    def _del(self, k):
        if self.__n == 0:
            return None
        self._swap(k, self.__n)
        popped = self.__heap.pop(-1)
        self.__n -= 1
        self._down(k)
        return popped

    def pop(self):
        return self._del(1)

    def remove(self, val):
        for i in range(1, self.__n + 1):
            if self.comp(self.__heap[i], val) == 0:
                return self._del(i)

    def _cleanup(self):
        self.__heap = [None]
        self.__n = 0

    def __iter__(self):
        if self.size > 0:
            n = self.size if self.size < self.__n else self.__n
        else:
            n = self.__n
        for i in range(n):
            yield self.pop()
        self._cleanup()

    @property
    def empty(self):
        return True if len(self.__heap) == 1 and self.__n == 0 else False

    @property
    def count(self):
        return self.__n

    @property
    def sorted_list(self):
        return [x for x in self]
